
#include "bess.h"
#include "bessstubs.h"
#include "bess.pt"
#include "smesh.h"
#include "bess2vdt.h"
#include "fct2vdt.h"
#include "getoptP.h"

int
report_error (char *file, int line, char *msg)
{
  fprintf (stderr, "\n*** ERROR: %s, l.%d: %s\n", file, line, msg);
  exit (1);
  return FALSE;
}

int
report_warning (char *file, int line, char *msg)
{
  char buf[1024];
  sprintf (buf, "*** WARNING: %s, l.%d: %s\n", file, line, msg);
  fprintf (stderr, "\n%s\n", buf);
  return TRUE;
}

#define errexit(msg) report_error (__FILE__, __LINE__, msg)
#define warn(msg)    report_warning (__FILE__, __LINE__, msg)

static smesh_triangulation_p mesh = NULL;
static bess_topology_p bess = NULL;
static double crease_angle = 60;
static double corner_angle = 45;
static char *in = NULL, *out = NULL;
static char in_prefix[] = "../fct/";
static char in_suffix[] = ".fct";
static char out_suffix[] = ".vin";
static double mesh_size = 1;

static void
make_polys (MESHP mesh, int nvert, CkitPoint3 *poly)
{
  CkitPoint3 points[3];

  if        (nvert == 3) {
    if (!smesh_add_poly (mesh, poly, NULL))
      errexit ("Failed to add poly");
  } else if (nvert == 4) {
    VECMAC_2_VEC (points[0], =, poly[0]);
    VECMAC_2_VEC (points[1], =, poly[1]);
    VECMAC_2_VEC (points[2], =, poly[2]);
    if (!smesh_add_poly (mesh, points, NULL))
      errexit ("Failed to add poly");
    VECMAC_2_VEC (points[0], =, poly[2]);
    VECMAC_2_VEC (points[1], =, poly[3]);
    VECMAC_2_VEC (points[2], =, poly[0]);
    if (!smesh_add_poly (mesh, points, NULL))
      errexit ("Failed to add poly");
  }  else {
    char buf[128];
    sprintf (buf, "Cannot handle poly for nvert = %d\n", nvert);
    warn (buf);
  }
}


int
read_from_fct (MESHP mesh, char *dxf_file_name)
{
  const char *buf = NULL, *tok = NULL;
  int readon = TRUE;
  tokens_parser parser = NULL;
  int npt = 0, tpt = 0;
  CkitPoint3 *pt = NULL;
  FILE *handle = NULL;
  ckit_iobuf_t IOB = NULL;

#define CHKBUF                                          \
        {                                               \
          if (buf == NULL) {                            \
            fprintf (stderr,                            \
                     "Expected data, l.%d\n",           \
                     ckit_io_line_number (IOB));        \
            goto exit;                                  \
          }                                             \
        }
#define GETBUF                                          \
        {                                               \
          buf = ckit_io_scan_line (IOB, FALSE);         \
          CHKBUF;                                       \
          tokens_parse_line (parser, (char *)buf);      \
        }
#define REALLOC_PT_BUF(npt)                             \
  {                                                     \
    if (!pt || npt+1 > tpt) {                           \
      tpt = 2*npt + 1;                                  \
      pt = NEWP_REALLOC_OBJ_ARRAY (pt, tpt, CkitPoint3);     \
      if (!pt) {                                        \
        fprintf (stderr, "Not enough memory\n");        \
        goto exit;                                      \
      }                                                 \
    }                                                   \
  }
  
  handle = ckit_io_open_stream (dxf_file_name, "r");
  IOB    = ckit_io_new_buffer (handle);

  if (parser == NULL) {
    parser = tokens_new_parser ();
  }

  GETBUF;
  if (tokens_type_of_token (parser, 1) == TOKENS_STRING) {
    tok = tokens_token_as_string (parser, 1);
    if        (!tokens_token_and_keyword_equiv (tok, "PPM_START")) {
      errexit ("Expected \"PPM_START\"");
      goto exit;
    }
  } else {
    fprintf (stderr, "Expected \"PPM_START\", l.%d, got \"%s\"\n",  
             ckit_io_line_number (IOB),                         
             ckit_io_current_line (IOB, FALSE));
  }

  tpt = 0; pt  = NULL;

  while (readon) {
    GETBUF;
    if (tokens_type_of_token (parser, 1) == TOKENS_STRING) {
      tok = tokens_token_as_string (parser, 1);
      if        (tokens_token_and_keyword_equiv (tok, "POLY_START")) {
        npt = 0;
        while (1) {
          GETBUF;
          if (tokens_type_of_token (parser, 1) == TOKENS_STRING) {
            tok = tokens_token_as_string (parser, 1);
            if        (tokens_token_and_keyword_equiv (tok, "POSITION:")) {
              REALLOC_PT_BUF (npt);
              if (  (tokens_total_of_tokens (parser) != 4)
                  || tokens_type_of_token (parser, 2) != TOKENS_NUMBER
                  || tokens_type_of_token (parser, 3) != TOKENS_NUMBER
                  || tokens_type_of_token (parser, 4) != TOKENS_NUMBER
                  ) {
                fprintf (stderr,
                         "Expected three numbers, l.%d, got \"%s\"\n",  
                         ckit_io_line_number (IOB),                         
                         ckit_io_current_line (IOB, FALSE));
                goto exit;
              }
              pt[npt].x = tokens_token_as_double (parser, 2);
              pt[npt].y = tokens_token_as_double (parser, 3);
              pt[npt].z = tokens_token_as_double (parser, 4);
              npt++;
            } else if (tokens_token_and_keyword_equiv (tok, "POLY_END")) {
              make_polys (mesh, npt, pt);
              break;
            } else {
              fprintf (stderr, "Expected \"POLY_END\" or \"POSITION:\","
                       "l.%d, got \"%s\"\n",  
                       ckit_io_line_number (IOB),                         
                       ckit_io_current_line (IOB, FALSE));
            }
          } else {
            fprintf (stderr, "Expected \"POLY_END\" or \"POSITION:\","
                     "l.%d, got \"%s\"\n",  
                     ckit_io_line_number (IOB),                         
                     ckit_io_current_line (IOB, FALSE));
            goto exit;
          }
        } /* while */
      } else if (tokens_token_and_keyword_equiv (tok, "PPM_END")) {
        readon = FALSE;
      } else {
        fprintf (stderr, "Unknown input, l.%d, got \"%s\"\n",  
                     ckit_io_line_number (IOB),                         
                     ckit_io_current_line (IOB, FALSE));
      }
    }
  } /* while (readon) */
  
 exit:
  ckit_io_delete_buffer (IOB);
  fclose(handle);

  return TRUE;
}



static int
has_suffix (char *f, char *s /* suffix */)
{
  int lf, ls, i;

  if (!f || !s)
    return FALSE;
  
  lf = strlen((f)); ls = min(strlen((s)), lf);
  for (i = 0; i < ls; i++)
    if (toupper((f)[lf-i]) != toupper((s)[ls-i]))
      return FALSE;
  return TRUE;
}


static void 
usage (int argc, char **argv)
{
  char *option, *usage;

  fprintf (stderr, "usage: %s -i file + options below\n", argv[0]);
  
  if (ckit_opt_first_recognized (&option, &usage)) {
    fprintf (stderr, "%s %s\n",
             (option != NULL ? option : "?"),
             (usage  != NULL ? usage  : "?"));
  }
  while (ckit_opt_next_recognized (&option, &usage)) {
    fprintf (stderr, "%s %s\n",
             (option != NULL ? option : "?"),
             (usage  != NULL ? usage  : "?"));
  }
  exit (0);
}


void
strip_path (char *dest, char *src)
{
  char *p;
  
  p = strrchr(src, '/');
  if (p == NULL) sprintf(dest, "%s", src);
  else {
    sprintf(dest, "%s", ++p);
  }
}

static char *
gen_in (void)
{
  char *base, *testname;
  int il, pl, sl;
  
  if (!ckit_io_file_exists (in)) {
    il = strlen (in);
    pl = strlen (in_prefix);
    sl = strlen (in_suffix);
    base = NEWP_OBJ_ARRAY (il+1, char);
    strip_path (base, in);

    if (!has_suffix (base, in_suffix)) {
      testname = NEWP_OBJ_ARRAY (il+sl+1, char);
      sprintf (testname, "%s%s", base, in_suffix);
      if (ckit_io_file_exists (testname)) {
        NEWP_FREE_OBJ (base);
        NEWP_FREE_OBJ (in);
        in = testname;
        return in;
      }
      NEWP_FREE_OBJ (testname);
    }

    testname = NEWP_OBJ_ARRAY (pl+il+1, char);
    sprintf (testname, "%s%s", in_prefix, base);
    if (ckit_io_file_exists (testname)) {
      NEWP_FREE_OBJ (base);
      NEWP_FREE_OBJ (in);
      in = testname;
      return in;
    }
    NEWP_FREE_OBJ (testname);
    
    if (!has_suffix (base, in_suffix)) {
      testname = NEWP_OBJ_ARRAY (pl+sl+il+1, char);
      sprintf (testname, "%s%s%s", in_prefix, base, in_suffix);
      if (ckit_io_file_exists (testname)) {
        NEWP_FREE_OBJ (base);
        NEWP_FREE_OBJ (in);
        in = testname;
        return in;
      }
      NEWP_FREE_OBJ (testname);
    }
    
    NEWP_FREE_OBJ (base);
  }
  
  errexit ("Input file not found");
  return NULL;
}



static char *
gen_out (void)
{
  char *base, *testname;
  int sl;
  
  base = NEWP_OBJ_ARRAY (strlen (out)+1, char);
  strip_path (base, out);
  
  if (has_suffix (base, in_suffix)) {
    base[strlen(base) - strlen(in_suffix)] = '\0';
  }
  
  sl = max (strlen (in_suffix), strlen (out_suffix));

  testname = NEWP_OBJ_ARRAY (strlen(base)+sl+1, char);
  sprintf (testname, "%s%s", base, out_suffix);
  if (ckit_io_file_exists (testname)) {
    char *buf = NEWP_OBJ_ARRAY (strlen (testname) + 128, char);
    sprintf (buf, "Output file \"%s\" exists; cannot overwrite", testname);
    errexit (buf);
  } else {
    NEWP_DUPL_STRING (out, testname, { errexit ("Not enough memory"); });
    NEWP_FREE_OBJ (base);
    return out;
  }
  return NULL;
}


static void 
process_options (int argc, char **argv)
{
  int help = FALSE;
  int report_opts = FALSE;

#define DOSTRINGZ(s) #s
#define STRINGZ(s) DOSTRINGZ(s)
#define SET_BOOL(name)                                          \
  {                                                             \
    if (report_opts) { fprintf (stderr, "Set: %s\n", #name); }  \
    name = TRUE;                                                \
  }
#define SET_WITH_ARG(name, format)                                        \
  {                                                                       \
    if (report_opts) { fprintf (stderr, "Set: %s = %" #format "\n",       \
                                STRINGZ(name), name); }                   \
  }
  
  GET_OPT_BOOL (argc, argv, "-report_opts", SET_BOOL(report_opts),
                "prints options as they are being set");
  
  GET_OPT_BOOL (argc, argv, "-help", SET_BOOL(help),
                "prints usage for each option recognized");
  
  GET_OPT_ARG (argc, argv, "-i", char, in,
               SET_WITH_ARG (in, s), "input file");
  
  GET_OPT_ARG (argc, argv, "-o", char, out,
               SET_WITH_ARG (out, s), "output file");
  
  GET_OPT_ARG (argc, argv, "-crease_angle", double, crease_angle,
               SET_WITH_ARG (crease_angle, g), "crease angle (degrees)");
  
  GET_OPT_ARG (argc, argv, "-corner_angle", double, corner_angle,
               SET_WITH_ARG (corner_angle, g), "corner angle (degrees)");
  
  if (help)
    usage (argc, argv);

  if (!in)
    usage (argc, argv);

  in  = gen_in ();
  if (!out) {
    NEWP_DUPL_STRING (out, in, { errexit ("Not enough memory"); });
  }
  out = gen_out ();
}

static double 
compute_mesh_size (BESSP bess)
{
  double el, avgel = 0;
  int ne = 0;
  MEP me;
  LOC v, l[3];
  
  me = NULL;
  while ((me = bess_next_me (bess, me))) {
    bess_me_get_loc (me, l);
    VECMAC_3_VEC (v, =, l[1], -, l[0]);
    el = VECMAC_NORM (v);
    avgel += el;
    ne++;
  }
  if (ne > 0)
    return avgel / ne;
  else
    return mesh_size;
}

int 
main (int argc, char **argv)
{
  process_options (argc, argv);
  
  mesh = smesh_new_triangulation (NULL, 0);
  
  if (!read_from_fct (mesh, in)) {
    char *buf = NEWP_OBJ_ARRAY (strlen (in) + 128, char);
    sprintf (buf, "Failed to read \"%s\"", in);
    errexit (buf);
  }

  bess = bess_new ();
  {
    bess_opts_t opts;
    opts.option = BESS_CREASE_ANGLE;
    opts.value.d = crease_angle;
    bess_config (bess, &opts);
    
    opts.option = BESS_CORNER_ANGLE;
    opts.value.d = corner_angle;
    bess_config (bess, &opts);
  }
  bess_build (bess, mesh);
  
  {
    FILE *fp = fopen (out, "w");
    if (!fp) {
      char *buf = NEWP_OBJ_ARRAY (strlen (out) + 128, char);
      sprintf (buf, "Failed to open \"%s\"", out);
      errexit (buf);
    }
    fprintf (fp, "mesh_size %g\n", compute_mesh_size (bess));
    bess_to_vdt (bess, fp);
    fclose (fp);
  }
  return 0;
}
