
#include "gleb.h"

static int ev[3][2] = {{2,0}, {0,1}, {1,2}};
static int tv[3] = {1,2,0};

typedef struct quasi_priority_queue_t {
  double    max_h_h_hat_ratio;
  int       nbucket;
  LIST     *buckets;
}              quasi_priority_queue_t;
#define QPQP quasi_priority_queue_t *

typedef struct gleb_t {
  LIST                    input_triangles;
  LIST                    output_triangles;
  void                   *h_hat_func_data;
  gleb_h_hat_func         h_hat_func;
  quasi_priority_queue_t *qpq;
}              gleb_t;
#undef GLEB
#define GLEBP gleb_t *

typedef struct gleb_pt_t {
  CkitPoint3      loc;
  double          h_hat;
}              gleb_pt_t;

typedef struct gleb_tri_t {
  gleb_pt_t     pt[3];
}              gleb_tri_t;
#undef GTP
#define GTP gleb_tri_t *

static int
qpq_map (QPQP qpq, double h_h_hat)
{
  double r;
  r = (1 - h_h_hat / qpq->max_h_h_hat_ratio);
  if (r < 0) r = 0;
#if 0
  printf ("%g (%d)\n", r, (int) ((qpq->nbucket - 1) * r));
#endif
  return (qpq->nbucket - 1) * r;
}


static double
h_hat_along_edge (GTP t, int i)
{
  double h0    = t->pt[ev[i][0]].h_hat;
  double h1    = t->pt[ev[i][1]].h_hat;
  return 2 * (h0 * h1) / (h0 + h1);
}

static int
edge_h_h_hat (GTP t, double hhr[3])
{
  int mhhri;
  CkitPoint3 v;

  VECMAC_3_VEC (v, =, t->pt[0].loc, -, t->pt[2].loc);
  hhr[0] = VECMAC_NORM (v) / h_hat_along_edge (t, 0);
  VECMAC_3_VEC (v, =, t->pt[1].loc, -, t->pt[0].loc);
  hhr[1] = VECMAC_NORM (v) / h_hat_along_edge (t, 1);
  VECMAC_3_VEC (v, =, t->pt[2].loc, -, t->pt[1].loc);
  hhr[2] = VECMAC_NORM (v) / h_hat_along_edge (t, 2);

  if      (hhr[0] >= hhr[1] && hhr[0] >= hhr[2]) mhhri = 0;
  else if (hhr[1] >= hhr[0] && hhr[1] >= hhr[2]) mhhri = 1;
  else                                           mhhri = 2;
  return mhhri;
}



static QPQP 
new_qpq (void)
{
  quasi_priority_queue_t *qpq;
  qpq = NEWP_OBJ (quasi_priority_queue_t);
  if (!qpq) return NULL;
  qpq->max_h_h_hat_ratio = 1;
  qpq->nbucket           = 0;
  qpq->buckets           = NULL;
  return qpq;
}


static int
fill_qpq (GLEBP gleb)
{
  GTP t;
  double hhr[3];
  int mhhri;
  double h_hat;
  int i;
  double max_h_h_hat = 0;

  t = NULL;
  while ((t = (GTP) get_list_next (gleb->input_triangles, t))) {
    mhhri = edge_h_h_hat (t, hhr);
    max_h_h_hat = max (max_h_h_hat, hhr[mhhri]);
  }

  gleb->qpq->max_h_h_hat_ratio = max_h_h_hat;
  gleb->qpq->nbucket = (gleb->qpq->max_h_h_hat_ratio * 10);
  if (gleb->qpq->nbucket < 1) gleb->qpq->nbucket = 1;
  gleb->qpq->buckets = NEWP_OBJ_ARRAY (gleb->qpq->nbucket, LIST);
  if (!gleb->qpq->buckets) {
    abort ();
  }
  FOR (i, gleb->qpq->nbucket) {
    gleb->qpq->buckets[i] = make_list ();
  }
    
  t = NULL;
  while ((t = (GTP) get_list_next (gleb->input_triangles, t))) {
    mhhri  = edge_h_h_hat (t, hhr);
    i = qpq_map (gleb->qpq, hhr[mhhri]);
    add_to_tail (gleb->qpq->buckets[i], t);
  }

  flush_list (gleb->input_triangles, NOT_NODES);
  
  return TRUE;
}

GLEBP
gleb_new (gleb_h_hat_func h_hat_func, void *h_hat_func_data)
{
  GLEBP gleb;
  gleb = NEWP_OBJ (gleb_t);
  if (!gleb) return NULL;

  gleb->input_triangles     = make_list ();
  gleb->output_triangles    = make_list ();
  gleb->h_hat_func_data     = h_hat_func_data;
  gleb->h_hat_func          = h_hat_func;
  gleb->qpq                 = new_qpq ();
  if (!gleb->qpq) return NULL;
  
  return gleb;
}

static GTP
new_t (CkitPoint3 pts[3], double h_hat[3])
{
  GTP t;
  int i;
  
  t = NEWP_OBJ (gleb_tri_t);
  if (!t) return NULL;

  FOR (i, 3) {
    t->pt[i].loc = pts[i];
    t->pt[i].h_hat = h_hat[i];
  }

  return t;
}

int
gleb_add (GLEBP gleb, CkitPoint3 pts[3])
{
  GTP t;
  double h_hat[3];
  int i;
  
  FOR (i, 3) {
    h_hat[i] = gleb->h_hat_func (gleb->h_hat_func_data,
                                 pts[i].x, pts[i].y, pts[i].z);
  }
  t = new_t (pts, h_hat);
  if (!t) return FALSE;
  add_to_tail (gleb->input_triangles, t);

  return TRUE;
}


static GTP
split_t (GTP t, int meli)
{
  CkitPoint3 pts[3];
  CkitPoint3 v;
  double el[3];
  GTP newt = NULL;
  double h_hat[3], hv_hat;
  
  VECMAC_3_VEC (v, =, 0.5 * t->pt[ev[meli][0]].loc, 
                   +, 0.5 * t->pt[ev[meli][1]].loc);
  hv_hat = 0.5 * (t->pt[ev[meli][0]].h_hat + t->pt[ev[meli][1]].h_hat);
  
  pts[0] = t->pt[tv[meli]].loc;    h_hat[0] = t->pt[tv[meli]].h_hat;
  pts[1] = t->pt[ev[meli][0]].loc; h_hat[1] = t->pt[ev[meli][0]].h_hat;
  pts[2] = v;                      h_hat[2] = hv_hat;
  newt   = new_t (pts, h_hat);
  pts[0] = t->pt[tv[meli]].loc;    h_hat[0] = t->pt[tv[meli]].h_hat;
  pts[1] = v;                      h_hat[1] = hv_hat;
  pts[2] = t->pt[ev[meli][1]].loc; h_hat[2] = t->pt[ev[meli][1]].h_hat;
  t->pt[0].loc = pts[0]; t->pt[0].h_hat = h_hat[0];
  t->pt[1].loc = pts[1]; t->pt[1].h_hat = h_hat[1];
  t->pt[2].loc = pts[2]; t->pt[2].h_hat = h_hat[2];

  return newt;
}

static void
setup_output (GLEBP gleb)
{
  int i, nbucket = gleb->qpq->nbucket;
  LIST *buckets = gleb->qpq->buckets;
  GTP t;
  
  flush_list (gleb->output_triangles, NOT_NODES);

  FOR (i, nbucket) {
    if (list_length (buckets[i]) > 0) {
      while ((t = (GTP) get_from_head (buckets[i]))) {
        add_to_tail (gleb->output_triangles, t);
      }
    }
  }
}

int
gleb_split (GLEBP gleb)
{
  int i, j;
  QPQP qpq;
  int nbucket;
  LIST *buckets;
  int anychange;
  double hhr[3];
  int mhhri;
  GTP t;
  GTP newt;

  if (!fill_qpq (gleb)) return FALSE;
  
  qpq     = gleb->qpq;
  nbucket = qpq->nbucket;
  buckets = qpq->buckets;
  
  do {
  restart:
    anychange = FALSE;
    FOR (i, nbucket) {
      if (list_length (buckets[i]) > 0) {
        t = NULL;
        while ((t = (GTP) get_list_next (buckets[i], t))) {
          mhhri = edge_h_h_hat (t, hhr); 
         
          if (hhr[mhhri] >= 1) {
            get_from_current_prev (buckets[i]); /* remove t from bucket */

            newt = split_t (t, mhhri); /* new triangle */
            
            mhhri = edge_h_h_hat (t, hhr); /* re-insert t */
            j = qpq_map (qpq, hhr[mhhri]);
            add_to_tail (buckets[j], t);
            
            mhhri = edge_h_h_hat (newt, hhr); /* insert newt */
            j = qpq_map (qpq, hhr[mhhri]);
            add_to_tail (buckets[j], newt);
            
            anychange = TRUE;     /* yes, need to recurse */
            goto restart;         /* restart from the beginning */
          } /* if (el[meli] / h_hat > 1) */
        } /* while ((t = (GTP) get_list_next (buckets[i], t)))   */
      } /* if (list_length (buckets[i]) */
    } /* FOR (i, nbucket) */
  } while (anychange);

  setup_output (gleb);
  
  return TRUE;
}

void *
gleb_next (GLEBP gleb, CkitPoint3 pts[3], void *prev)
{
  GTP t;
  t = (GTP) get_list_next (gleb->output_triangles, prev);
  if (t) {
    pts[0] = t->pt[0].loc;
    pts[1] = t->pt[1].loc;
    pts[2] = t->pt[2].loc;
  }
  return t;
}


void
gleb_free (GLEBP gleb)
{
  int i;
  GTP t;
  
  if (gleb->qpq) {
    if (gleb->qpq->buckets) {
      FOR (i, gleb->qpq->nbucket) {
        if (gleb->qpq->buckets[i]) {
          free_list (gleb->qpq->buckets[i], NOT_NODES);
        }
      }
      NEWP_FREE_OBJ (gleb->qpq->buckets);
    }
    NEWP_FREE_OBJ (gleb->qpq);
  }
  free_list (gleb->input_triangles, AND_NODES);

  while ((t = (GTP) get_from_head (gleb->output_triangles))) {
    NEWP_FREE_OBJ (t);
  }
  free_list (gleb->output_triangles, NOT_NODES);

  NEWP_FREE_OBJ (gleb);
}
