
#include "bess.h"
#include "bessstubs.h"
#include "bess.pt"
#include "mic.h"

typedef struct edge_isect_t {
  MEP    me;
  double lambda;                /* parameter of G3D_intersect_tri_and_seg() */
}              edge_isect_t;

typedef struct mi_checker_t {
  BESSP                     bess;
  dict_dictionary           interfering_triangles;
  dict_dictionary           interfering_edges;
  lobs_3d_front_end         tlobs;
}              mi_checker_t;



static void
t_get_bbox (MFP mf, G3D_box_t *box)
{
  LOC loc[3];
  bess_mf_get_loc (mf, loc);
  G3D_INIT_BOX ((*box), loc[0]);
  G3D_UPDT_BOX ((*box), loc[1]);
  G3D_UPDT_BOX ((*box), loc[2]);
}


static void
add_t_to_lobs (mi_checker_p mic)
{
  MFP mf;
  G3D_box_t box;
  
  mf = NULL;
  while ((mf = bess_next_mf (mic->bess, mf))) {
    t_get_bbox (mf, &box);
    lobs_3d_add_object (mic->tlobs, mf,
                        box.lower.x, box.lower.y, box.lower.z,
                        box.upper.x, box.upper.y, box.upper.z);
  }
}


static int
v_shares_e (MVP v, MEP e)
{
  MEP ome;
  ome = NULL;
  while ((ome = bess_mv_get_next_me (v, ome))) {
    if (ome == e) {
      return TRUE;
    }
  }
  return FALSE;
}

static int
seg_tri_intersect (LOC el[2], LOC fl[3], double *lambda)
{
  G3D_loc_t s, P, V;
  G3D_tri_t t;
  G3D_box_t b;
  int i;
  double eps;

  P = el[0];
  V = el[1];
  VECMAC_2_VEC (V, -=, P);
                   
  FOR (i, 3) { t.loc[i] = fl[i]; }
  
  G3D_INIT_BOX (b, t.loc[0]);
  G3D_UPDT_BOX (b, t.loc[1]);
  G3D_UPDT_BOX (b, t.loc[2]);
  eps = (G3D_BOX_RANGE (b, x) + G3D_BOX_RANGE (b, y) + G3D_BOX_RANGE (b, z)) / 10000;
  
  return (G3D_intersect_tri_and_seg (&s, lambda, &P, &V, &t, eps));
}

static int
have_mf (mi_checker_p mic, MFP mf)
{
  return ((MFP) dict_find_obj_with_handle_name (mic->interfering_triangles, mf) == mf);
}

static void
cond_add_mf (mi_checker_p mic, MFP mf)
{
  if (have_mf (mic, mf)) return;
  else                   dict_add_obj_handle_name (mic->interfering_triangles, mf, mf);
}

static int
have_me (mi_checker_p mic, MEP me, double *lambda)
{
  *lambda=0;
  edge_isect_t *eis = 
    ((edge_isect_t *) dict_find_obj_with_handle_name (mic->interfering_edges, me));
  if (eis && (eis->me == me)) {
    *lambda = eis->lambda;
    return TRUE;
  }
  return FALSE;
}

static void
cond_add_me (mi_checker_p mic, MEP me, double lambda)
{
  double olambda;
  if (have_me (mic, me, &olambda)) return;
  else {
    edge_isect_t *eis = NEWP_OBJ(edge_isect_t);
    eis->me = me;
    eis->lambda = lambda;
    dict_add_obj_handle_name (mic->interfering_edges, eis, me);
  }
}

static void
intersect_t (mi_checker_p mic, MEP me, MFP mf)
{
  LOC el[2], fl[3];
  int i;
  MFP omf;
  MVP fmv[3];
  double lambda=0;
  
  bess_mf_get_loc (mf, fl);
  bess_me_get_loc (me, el);

  if (!seg_tri_intersect (el, fl, &lambda)) return;

  bess_mf_get_mv (mf, fmv);
  
  FOR (i, 3) {
    if (v_shares_e (fmv[i], me)) {
      /* detected intersection with an edge passing through the vertex;
         ignore the intersection
         */
      return;
    }
  }
  /* now we've got genuine intersection: add to dictionaries */
  cond_add_me (mic, me, lambda);
  omf = NULL;
  while ((omf = bess_me_get_next_mf (me, omf))) {
    cond_add_mf (mic, omf);
  }
  cond_add_mf (mic, mf);
}

static void
intersect_with_triangles (mi_checker_p mic, MEP me)
{
  LOC l[2];
  G3D_box_t box;
  MFP mf;

  bess_me_get_loc (me, l);
  G3D_INIT_BOX (box, l[0]);
  G3D_UPDT_BOX (box, l[1]);
  
  lobs_3d_search (mic->tlobs,
                  box.lower.x, box.lower.y, box.lower.z,
                  box.upper.x, box.upper.y, box.upper.z);

  mf = (MFP) lobs_3d_first_overlapping_object (mic->tlobs);
  while (mf) {
    intersect_t (mic, me, mf);
    mf = (MFP) lobs_3d_next_overlapping_object  (mic->tlobs);
  }
}


static void
interference_check (mi_checker_p mic)
{
  MEP me;
  me = NULL;
  while ((me = bess_next_me (mic->bess, me))) {
    intersect_with_triangles (mic, me);
  }
}



mi_checker_p
mi_create_checker (BESSP bess)
{
  mi_checker_p mi;
  mi = NEWP_OBJ (mi_checker_t);
  mi->bess                  = bess;
  mi->interfering_triangles = dict_create_dictionary ();
  mi->interfering_edges     = dict_create_dictionary ();
  mi->tlobs                 = lobs_3d_new_instance (NULL, 100);
  return mi;
}


void
mi_destroy_checker (mi_checker_p mic)
{
/* RAW */
}

int
mi_check (mi_checker_p mic)
{
  add_t_to_lobs (mic);
  interference_check (mic);
  return 1;
}

int
mi_does_interfere (mi_checker_p mic, void  *mf)
{
  if (mic)
    return have_mf (mic, mf);
  else
    return FALSE;
}

int
mi_does_intersect (mi_checker_p mic, void  *me, double *lambda)
{
  if (mic)
    return have_me (mic, me, lambda);
  else
    return FALSE;
}
