
#include "smesh.h"
#include <float.h>

#undef LOC
#define LOC      CkitPoint3

#undef MESH
#define MESH     smesh_triangulation_t
#undef MESHP
#define MESHP    smesh_triangulation_p
#undef MF
#define MF       smesh_triangle_t
#undef ME
#define ME       smesh_edge_t
#undef MV
#define MV       smesh_vertex_t
#undef MFP
#define MFP      smesh_triangle_p
#undef MEP
#define MEP      smesh_edge_p
#undef MVP
#define MVP      smesh_vertex_p
#define MRID     void *


#ifndef FOR
#  define FOR(i_, n_) for (i_ = 0; i_ < n_; i_++)
#endif

#ifndef min
#   define min(A,B) (((A) < (B)) ? (A) : (B))
#endif
#ifndef max
#   define max(A,B) (((A) > (B)) ? (A) : (B))
#endif

#define REPERR(st, err)                                 \
  {                                                     \
    if (st->error_notify) {                             \
      (*st->error_notify) (st, __FILE__, __LINE__, err);\
    } else {                                            \
      fprintf (stderr, "\nError (%s, l.%d): %s\n",      \
               __FILE__, __LINE__, err);                \
    }                                                   \
  }


#undef BOOL
typedef char BOOL;

typedef struct MF {
  int         id;
  struct ME  *e[3];
  void       *ap;
  BOOL        oriented;
  void       *surf;
}              MF;

typedef struct ME {
  int                    id;
  MVP                    v[2];
  LIST_OF (struct MF *)  tl;    /* list of incident triangles.
                                   Manifold surface should have at most two, at least one;
                                   non-manifold surface may have at least one, without an upper limit
                                   */
  void       *ap;
}              ME;

typedef struct MV {
  int                   id;
  LOC                   loc;
  LIST_OF (struct ME *) el;
  void       *ap;
}              MV;

typedef struct MESH {
  smesh_error_notify_func   error_notify;
  LIST_OF (struct MV *)     vl;
  LIST_OF (struct ME *)     el;
  LIST_OF (struct MF *)     tl;
  pirat_front_end_t        *vp;
  dict_dictionary           vd;
  double                    vltol;
  G3D_box_t                 bbox;
  BOOL                      orientable;
}              MESH;


/*
   Create a new triangulation.
 */
MESHP /* triangulation */
smesh_new_triangulation (
                         smesh_error_notify_func error_notify,
                         double vltol
                         )
{
  MESHP st;
  LOC zero = {0,0,0};

  st = NEWP_OBJ (MESH);
  if (!st) return NULL;

  st->vl         = make_list ();
  st->el         = make_list ();
  st->tl         = make_list ();
  st->vp         = pirat_new_pirat (100, 1000 * DBL_EPSILON);
  st->vd         = dict_create_dictionary ();
  st->vltol      = vltol;
  st->bbox.lower = zero;
  st->bbox.upper = zero;
  st->orientable = TRUE;
  
  return st;
}


static void
attach_v (MESHP st, MVP v)
{
  add_to_tail (st->vl, v);
  pirat_add_pt (st->vp, v, &v->loc);
  dict_add_obj_int_name (st->vd, v, v->id);
}

static MV *
new_v (int vid, LOC *l)
{
  MV *v;
  static int id = 0;
  
  v = NEWP_OBJ (MV);
  if (!v) return NULL;
  v->id  = (vid >= 0 ? vid : ++id);
  v->loc = *l;
  v->el  = make_list ();
  v->ap  = NULL;
  return v;
}

static void
v_attach_e (MVP v, MEP e)
{
  add_to_tail (v->el, e);
}

static double
e_length_squared (MEP e)
{
  CkitPoint3 vec;
  VECMAC_3_VEC (vec, =, e->v[0]->loc, -, e->v[1]->loc);
  return VECMAC_DOT (vec, vec);
}

static double
shortest_e_length_at_v (MVP v)
{
  if (list_empty (v->el)) return 0;
  else {
    double fsel = FLT_MAX, l;
    MEP e;
    e = NULL;
    while ((e = (MEP) get_list_next (v->el, e))) {
      l = e_length_squared (e);
      fsel = min (fsel, l);
    }
    return sqrt (fsel);
  }
}

static MV *
find_v (MESHP st, LOC *l)
{
  MV *v;
  G3D_box_t range;
  double vltol = st->vltol;
  double vltol_bound = st->vltol;
  const int max_nretries = 5;
  int nretries = 0;
  double fract = 10;

 repeat_search:
  
  G3D_INIT_BOX (range, (*l));
  G3D_INFLATE_BOX (range, vltol);
  
  pirat_search_range (st->vp, &range, INT_MAX);
  v = (MV *) pirat_first_in_range (st->vp);
  if (v  && st->vp->num_of_points_found > 1) {
    if (++nretries < max_nretries) {
      double sel = shortest_e_length_at_v (v);
      if (sel > 0) {
        vltol = min (vltol, sel / fract);
        vltol = min (vltol, vltol_bound);
        fract *= 10;
        goto repeat_search;
      }
    }
    if (pirat_next_in_range (st->vp)) {
      char buf[512];
      sprintf (buf, "Two or more vertices at a location {%g, %g, %g}",
               l->x, l->y, l->z);
      REPERR (st, buf);
    }
  }
  
  return v;
}



static ME *
new_e (MESHP st, MV *sv, MV *ev)
{
  ME *e;
  static int id = 0;

  e = NEWP_OBJ (ME);
  if (!e) return NULL;
  e->id = ++id;
  e->v[0] = sv; v_attach_e (sv, e);
  e->v[1] = ev; v_attach_e (ev, e);
  e->tl   = make_list ();
  e->ap   = NULL;
  return e;
}

static MVP 
get_existing_v (MESHP st, LOC *l)
{
  int vll;
  if ((vll = list_length (st->vl)) < 1) {
    return NULL;
  } else {
    G3D_UPDT_BOX (st->bbox, (*l));
    if (st->vltol <= 0) {
      st->vltol = ( 0.00001
                   / pow ((double)vll, 0.3333)
                   * (G3D_BOX_RANGE (st->bbox, x) +
                      G3D_BOX_RANGE (st->bbox, y) +
                      G3D_BOX_RANGE (st->bbox, z))
                   );
    }
    return find_v (st, l);
  }
}

static MV *
merge_v (MESHP st, int id, LOC *l)
{
  MV *v;
  int vll = list_length (st->vl);
  
  if (vll < 1) {
    G3D_INIT_BOX (st->bbox, (*l));
    v = new_v (id, l);
    attach_v (st, v);
    return v;
  } else {
    v = get_existing_v (st, l);
    if (!v) {
      v = new_v (id, l);
      attach_v (st, v);
    } 
    return v;
  }
}


static ME *
conn_e (MV *v, MV *nv)
{
  ME *e;
  loop_through_list (v->el, e, ME *) {
    if      (e->v[0] == v && e->v[1] == nv)
      return e;
    else if (e->v[0] == nv && e->v[1] == v)
      return e;
  }
  return NULL;
}

static ME *
merge_e (MESHP st, MV *sv, MV *ev)
{
  ME *e;
  e = conn_e (sv, ev);
  if (!e) {
    e = new_e (st, sv, ev);
    add_to_tail (st->el, e);
  }
  return e;
}


#define T_USES_E(t, ed) (t->e[0] == ed || t->e[1] == ed || t->e[2] == ed)

static MF *
find_t (MESHP st, ME *e0, ME *e1, ME *e2)
{
  MF *t, *nt;
  loop_through_list (e0->tl, t, MF *) {
    if (T_USES_E (t, e1)) {
      loop_through_list (e1->tl, nt, MF *) {
        if (T_USES_E (nt, e2)) {
          return nt;
        }
      }
    }
  }
  return NULL;
}

static void
e_attach_t (ME *e, MF *t)
{
  add_to_tail (e->tl, t);
}

static MF *
new_t (ME *e0, ME *e1, ME *e2, void *surf)
{
  MF *t;
  static int id = 0;
  
  t = NEWP_OBJ (MF);
  if (!t) return NULL;
  t->id  = ++id;
  t->e[0]     = e0; e_attach_t (t->e[0], t);
  t->e[1]     = e1; e_attach_t (t->e[1], t);
  t->e[2]     = e2; e_attach_t (t->e[2], t);
  t->oriented = FALSE;
  t->surf     = surf;
  return t;
}

static MF *
merge_t (MESHP st, ME *e0, ME *e1, ME *e2, void *surf)
{
  MF *t;
  t = new_t (e0, e1, e2, surf);
  add_to_tail (st->tl, t);
  return t;
}

static BOOL
has_poly (MESHP st,
          LOC *locations /* vertex locations */,
          void *surf
          )
{
  MV *v[3];
  MEP e[3];
  MFP t;
  int i;
  FOR (i, 3) {
    v[i] = get_existing_v (st, &locations[i]);
    if (v[i] == NULL) return FALSE; /* can't have the poly; don't even have the vertex */
  }
  /* now I know I have the vertices, check if I have the edges */
  FOR (i, 3) {
    e[i] = conn_e (v[i], v[(i+1)%3]);
    if (e[i] == NULL) return FALSE; /* can't have the poly; don't even have the edge */
  }
  /* now I know I have the edges, check if I have the poly */
  t = find_t (st, e[0], e[1], e[2]);
  return (t != NULL);
}

static MFP
add_poly (MESHP st,
          LOC *locations /* vertex locations */,
          void *surf
          )
{
  MV *v[3];
  ME *e[3];
  MF *t;
  int i;

  FOR (i, 3) {
    v[i] = get_existing_v (st, &locations[i]);
  }

  if ((v[0] != NULL && (v[0] == v[1])) ||
      (v[1] != NULL && (v[1] == v[2])) ||
      (v[2] != NULL && (v[2] == v[0]))) {
    REPERR (st, "Triangle with 2 identical vertices");
    return NULL;
  }

  FOR (i, 3) {
    if (v[i] == NULL) {
      v[i] = new_v (-1 /* new */, &locations[i]);
      attach_v (st, v[i]);
    }
  }
  
  e[0] = merge_e (st, v[2], v[0]);
  e[1] = merge_e (st, v[0], v[1]); 
  e[2] = merge_e (st, v[1], v[2]);
  FOR (i, 3) {
    if (!e[i]) {
      REPERR (st, "Failed to add edge");
      return NULL;
    }
  }
  t = merge_t (st, e[0], e[1], e[2], surf);
  if (!t) {
    REPERR (st, "Failed to add triangle");
    return NULL;
  }

  return t;
}


/*
   Add a triangle to a triangulation.
   The triangle must not be degenerate.
 */
int /* TRUE if successfully merged; FALSE otherwise */
smesh_add_poly (MESHP st,
                LOC *locations, /* vertex locations */
                void *surf
                )
{
  MF *t;
  if (has_poly (st, locations, surf))
    return TRUE;
  t = add_poly (st, locations, surf);
  return (t != NULL);
}


/*
   Add a triangle to a triangulation.
   The triangle must not be degenerate.
 */
int /* TRUE if successfully merged; FALSE otherwise */
smesh_add_poly_with_ap (MESHP st,
                        LOC *locations, /* vertex locations */
                        void *ap, /* associated pointer */
                        void *surf
                        )
{
  MF *t;
  if (has_poly (st, locations, surf))
    return TRUE;
  t = add_poly (st, locations, surf);
  if (t) {
    smesh_t_set_ap (t, ap);
    return TRUE;
  } else {
    return FALSE;
  }
}



int
smesh_add_v (MESHP st, int id, LOC *loc)
{
  MV *v;
  v = merge_v (st, id, loc);
  if (!v) {
    REPERR (st, "Failed to add vertex");
    return FALSE;
  }
  return TRUE;
}

/*
   Add a triangle to a triangulation.
   The triangle must not be degenerate.
 */
int /* TRUE if successfully merged; FALSE otherwise */
smesh_add_t_with_ap (MESHP st,
                     int vid[3],
                     void *ap,
                     void *surf
                     )
{
  MV *v[3];
  ME *e[3];
  MF *t;
  int i;

  FOR (i, 3) {
    v[i] = dict_find_obj_with_int_name (st->vd, vid[i]);
    if (!v[i]) {
      REPERR (st, "Failed to add vertex");
      return FALSE;
    }
  }

  e[0] = merge_e (st, v[2], v[0]);
  e[1] = merge_e (st, v[0], v[1]); 
  e[2] = merge_e (st, v[1], v[2]);
  FOR (i, 3) {
    if (!e[i]) {
      REPERR (st, "Failed to add edge");
      return FALSE;
    }
  }
  t = merge_t (st, e[0], e[1], e[2], surf);
  if (!t) {
    REPERR (st, "Failed to add triangle");
    return FALSE;
  }
  smesh_t_set_ap (t, ap);

  return TRUE;
}


/*
   Iterate triangles.
 */
MFP /* NULL if there are no more triangles */
smesh_next_t (MESHP st, /* surface triangulation */
              MFP o /* start iteration by passing NULL */
              )
{
  return (MFP) get_list_next (st->tl, o);
}

/*
   Iterate edges.
 */
MEP /* NULL if there are no more edges */
smesh_next_e (MESHP st, /* surface triangulation */
              MEP o /* start iteration by passing NULL */
              )
{
  return (MEP) get_list_next (st->el, o);
}

/*
   Iterate vertices.
 */
MVP /* NULL if there are no more vertices */
smesh_next_v (MESHP st, /* surface triangulation */
              MVP o /* start iteration by passing NULL */
              )
{
  return (MVP) get_list_next (st->vl, o);
}

static MV *
common_v (ME *e, ME *oe)
{
  if      ((e->v[0] == oe->v[0]) || (e->v[0] == oe->v[1]))
    return e->v[0];
  else if ((e->v[1] == oe->v[0]) || (e->v[1] == oe->v[1]))
    return e->v[1];
  else
    return NULL;
}



/*
   Get triangle normal.
 */
int /* TRUE if normal is valid; FALSE otherwise */
smesh_t_get_normal (MFP t, /* triangle */
                    double xi, double eta, 
                    LOC *normal
                    )
{
  LOC n = {0,0,0}, f, s;
  MV *v[6];
  double nn;
  smesh_t_get_v (t, v);
  VECMAC_3_VEC (f, =, (v[1]->loc), -, (v[0]->loc));
  VECMAC_3_VEC (s, =, (v[2]->loc), -, (v[0]->loc));
  VECMAC_CROSS (n, =, f, x, s);
  nn = VECMAC_NORM (n);
  if (nn <= 0) return FALSE;
  VECMAC_NORMALIZE (n);
  *normal = n;
  return TRUE;
}


static void
destroy_v (MESHP st, MV *v)
{
  free_list (v->el, NOT_NODES);
  NEWP_FREE_OBJ (v);
}

static void
destroy_e (MESHP st, ME *e)
{
  free_list (e->tl, NOT_NODES);
  NEWP_FREE_OBJ (e);
}

static void
destroy_t (MESHP st, MF *t)
{
  NEWP_FREE_OBJ (t);
}

/*
   Destroy surface mesh.
 */
void
smesh_destroy_triangulation (MESHP st
                             )
{
  MV *v;
  ME *e;
  MF *t;

  loop_through_list (st->vl, v, MV *) {
    destroy_v (st, v);
  }
  loop_through_list (st->el, e, ME *) {
    destroy_e (st, e);
  }
  loop_through_list (st->tl, t, MF *) {
    destroy_t (st, t);
  }
  pirat_destroy_pirat (st->vp);
  dict_destroy_dictionary (st->vd);
  NEWP_FREE_OBJ (st);
}


static int
num_t_at_e (ME *e)
{
  return list_length (e->tl);
}

static int
manifold_e (ME *e)
{
  return (num_t_at_e (e) == 2);
}


static MF *
other_t (ME *e, MF *t)
{
  MF *ot;
  
  if (!manifold_e (e))
    return NULL; /* RAW: how do I handle non-manifold edges here? */

  loop_through_list (e->tl, ot, MF *) {
    if (ot != t)
      return ot;
  }
  return NULL;
}



static void
v_on_e_in_t (MFP mf, MEP me, MVP v[2])
{
  MVP tv[6];
  smesh_t_get_v (mf, tv);
  if        (me == mf->e[0]) {
    v[0] = tv[2];
    v[1] = tv[0];
  } else if (me == mf->e[1]) {
    v[0] = tv[0];
    v[1] = tv[1];
  } else if (me == mf->e[2]) {
    v[0] = tv[1];
    v[1] = tv[2];
  } else {
    v[0] = v[1] = NULL;
  }
}

static int
orient_t (MF *mf, MEP me, MFP omf)
{
  MVP mv[2], omv[2];

  v_on_e_in_t (mf, me, mv);
  v_on_e_in_t (omf, me, omv);
  if (mv[0] != omv[1] || mv[1] != omv[0]) {
    /* t is not oriented in the same manner as ot; swap two edges */
    ME *ae = omf->e[0]; omf->e[0] = omf->e[2]; omf->e[2] = ae;
  }
  return TRUE;
}


static int
orientation_is_compatible (MF *mf, MEP me, MFP omf)
{
  MVP mv[2], omv[2];

  v_on_e_in_t (mf, me, mv);
  v_on_e_in_t (omf, me, omv);
  return (mv[0] == omv[1] && mv[1] == omv[0]);
}

static void
orient_face (MESHP st, MFP smf)
{
  static STACK mfrs = NULL;
  MFP mf = smf;
  MFP omf;
  int i;
  LOC n;
  
  if (!mfrs) mfrs = make_stack ();
  if (!stack_empty (mfrs)) {
    REPERR (st, "stack is not empty");
  }
  
  push_stack (mfrs, mf);
  while (!stack_empty (mfrs)) {
    mf = (MFP) pop_stack (mfrs);
    mf->oriented = TRUE; 
    FOR (i, 3) {
      if (manifold_e (mf->e[i])) {
        omf = other_t (mf->e[i], mf);
        if (!omf->oriented) {
          orient_t (mf, mf->e[i], omf);
          smesh_t_get_normal (omf, 1/3, 1/3, &n);
          push_stack (mfrs, omf);
        } else {
          if (st->orientable) st->orientable = orientation_is_compatible (mf, mf->e[i], omf);
        }
      }
    } /* FOR (i, 3) */
  } /* while */
}

/*
   Orient surface mesh.
 */
int
smesh_orient (MESHP st)
{
  MF *t;
  int all_oriented;
  
  while (1) {
    all_oriented = TRUE;
    t = NULL;
    while ((t = (MF *) get_list_next (st->tl, t))) {
      if (!t->oriented) {
        orient_face (st, t);
        all_oriented = FALSE;
        break;
      }
    }
    if (all_oriented)
      return TRUE;
  }
}

/*
   Get edges.
 */
void
smesh_t_get_e (MFP t, /* triangle */
               MEP e[3] /* edges */
               )
{
  int i;
  FOR (i, 3) { e[i] = t->e[i]; }
}

/*
   Get associated pointer.
 */
void * /* pointer stored by the triangle */
smesh_t_get_ap (MFP t /* triangle */
                )
{
  return t->ap;
}

/*
   Set associated pointer.
 */
void * /* previously stored pointer */
smesh_t_set_ap (MFP t, /* triangle */
                void *ap /* pointer */
                )
{
  void *oap = t->ap;
  t->ap = ap;
  return oap;
}


/*
   Get surface tag.
 */
void * /* pointer stored by the triangle */
smesh_t_get_surf (MFP t /* triangle */
                )
{
  return t->surf;
}

/*
   Get id of mesh edge.
 */
int /* edge id */
smesh_e_get_id (MEP e /* edge */
                )
{
  return e->id;
}

/*
   Get associated pointer.
 */
void *
smesh_e_get_ap (MEP e)
{
  return e->ap;
}

/*
   Get incident vertices.
 */
int
smesh_e_get_v (MEP e, MVP v[3])
{
  v[0] = e->v[0];
  v[1] = e->v[1];
  return 2;
}

/*
   Iterate triangles incident to edge.
 */
MFP /* next triangle or NULL if there are no more */
smesh_e_get_next_t (MEP me,
                    MFP mf
                    )
{
  return (MFP) get_list_next (me->tl, mf);
}


/*
   Count incident triangles at edge.
 */
int
smesh_e_get_num_inc_t (MEP e)
{
  return list_length (e->tl);
}

/*
   Set associated pointer.
 */
void *
smesh_e_set_ap (MEP e,
                void *ap
                )
{
  void *oap = e->ap;
  e->ap = ap;
  return oap;
}

/*
   Get associated pointer.
 */
void *
smesh_v_get_ap (MVP v /* vertex */
                )
{
  return v->ap;
}

/*
   Set associated pointer.
 */
void *
smesh_v_set_ap (MVP v, void *ap)
{
  void *oap = v->ap;
  v->ap = ap;
  return oap;
}


/*
   Get vertex id
 */
int
smesh_v_get_id (MVP v)
{
  return v->id;
}

/*
   Get vertex location.
 */
void
smesh_v_get_loc (MVP v, /* vertex */
                 LOC *p /* location */
                 )
{
  *p = v->loc;
}

/*
   Set vertex location.
 */
void
smesh_v_set_loc (MVP v, /* vertex */
                 LOC *p /* location */
                 )
{
  v->loc = *p;
}

/*
   Iterate edges incident to vertex
 */
MEP
smesh_v_get_next_e (MVP v, MEP e)
{
  return (MEP) get_list_next (v->el, e);
}


/*
   Get triangle id
 */
int
smesh_t_get_id (MFP t)
{
  return t->id;
}

/*
   Get incident vertices.
 */
int
smesh_t_get_v (MFP t, /* triangle */
               MVP v[6] /* vertices in ccw order when
                         looking against the normal;
                         vertex 0 is shared by edges 0 and 1,
                         vertex 1 is shared by edges 1 and 2,
                         vertex 2 is shared by edges 2 and 0.
                         */
               )
{
  v[0] = common_v (t->e[0], t->e[1]);
  v[1] = common_v (t->e[1], t->e[2]);
  v[2] = common_v (t->e[2], t->e[0]);
  return 3;
}


/*
   Get regions on plus and minus side (plus side is where the normal points).
 */
int
smesh_t_get_regions (MFP mf, MRID *mridplus, MRID *mridminus)
{
  /* cannot answer region queries */
  return 0;
}


/*
   Get tangent vector to an edge at a vertex.
 */
void
smesh_e_get_tangent (MEP me, MVP mv, LOC *tangent)
{
  VECMAC_3_VEC ((*tangent),  =, me->v[1]->loc, -, me->v[0]->loc);
  VECMAC_NORMALIZE ((*tangent));
}


double
smesh_vltol (MESHP mesh)
{
  return mesh->vltol;
}

void
smesh_set_vltol (MESHP mesh, double vltol) { mesh->vltol = vltol; }

int
smesh_is_orientable (MESHP mesh) { return mesh->orientable; }

int
smesh_total_of_surf_triangles (MESHP mesh) { return list_length (mesh->tl); }

int
smesh_total_of_surf_edges (MESHP mesh) { return list_length (mesh->el); }

int
smesh_total_of_surf_vertices (MESHP mesh) { return list_length (mesh->vl); }
