
#include "tetmesh.h"

#undef LOC
#define LOC      CkitPoint3

#undef MESH
#define MESH     tet_triangulation_t
#undef MESHP
#define MESHP    tet_triangulation_p
#undef MT
#define MT       tet_tet_t
#undef MF
#define MF       tet_face_t
#undef ME
#define ME       tet_edge_t
#undef MV
#define MV       tet_vertex_t
#undef MTP
#define MTP      tet_tet_p
#undef MFP
#define MFP      tet_face_p
#undef MEP
#define MEP      tet_edge_p
#undef MVP
#define MVP      tet_vertex_p
#define MRID     void *


#ifndef FOR
#  define FOR(i_, n_) for (i_ = 0; i_ < n_; i_++)
#endif

#ifndef min
#   define min(A,B) (((A) < (B)) ? (A) : (B))
#endif
#ifndef max
#   define max(A,B) (((A) > (B)) ? (A) : (B))
#endif

#define REPERR(st, err)                                 \
  {                                                     \
    if (st->error_notify) {                             \
      (*st->error_notify) (st, __FILE__, __LINE__, err);\
    } else {                                            \
      fprintf (stderr, "\nError (%s, l.%d): %s\n",      \
               __FILE__, __LINE__, err);                \
      exit (1);                                         \
    }                                                   \
  }

#define CV_ON_F(facenum, vserialnum) g3dtet10_vertex_on_face[facenum][vserialnum*2]
static int e_serial_num_to_mev_serial_num[3] = {5, 1, 3};
#define MEV_ON_F(facenum, eserialnum) \
   g3dtet10_vertex_on_face[facenum][e_serial_num_to_mev_serial_num[eserialnum]]

#undef BOOL
typedef char BOOL;

typedef struct MV {
  int                   id;
  LOC                   loc;
  LIST_OF (MEP)         mel;
  void                 *ap;
}              MV;

typedef struct ME {
  int         id;
  MVP                    mv[3];
  LIST_OF (MFP)          mfl;
  void                  *ap;
}              ME;


typedef struct MF {
  int         id;
  MEP         me[3];
  MTP         mt[2]; /* mt[0] is on the side to which the normal points; mt[1] is on the other side */
  void       *ap;
}              MF;


typedef struct MT {
  int          id;
  int          rid; /* region id */
  MVP          mv[10];
  MTP          adjmt[4];
  int          new_rid;
  void        *ap;
}              MT;


typedef struct MESH {
  tet_error_notify_func   error_notify;
  int                     geom_degree;
  dict_dictionary         mvdict;
  dict_dictionary         mtdict;
  int                     num_mt;
  dict_dictionary         adjldict;
  LIST_OF (MEP)           mel;
  LIST_OF (MFP)           mfl;
  int                     invalid_rid;
  int                     curr_rid;
  int                     max_node_id;
  int                     orientable;
}              MESH;




#define MF_USES_ME(mf, e) (mf->me[0] == e || mf->me[1] == e || mf->me[2] == e)

static MF *
find_mf (MESHP st, ME *e0, ME *e1, ME *e2)
{
  MF *t, *nt;
  loop_through_list (e0->mfl, t, MF *) {
    if (MF_USES_ME (t, e1)) {
      loop_through_list (e1->mfl, nt, MF *) {
        if (MF_USES_ME (nt, e2)) {
          if (t != nt) {
            char buf[1024];
            sprintf(buf, "More than a single triangle use the same three edges: look around point %g %g %g",
                    e0->mv[0]->loc.x,
                    e0->mv[0]->loc.y,
                    e0->mv[0]->loc.z
                    );
            REPERR (st, buf);
          }
          return nt;
        }
      }
    }
  }
  return NULL;
}

static void
me_attach_mf (ME *e, MF *t)
{
  add_to_tail (e->mfl, t);
}

static MFP
new_mf (MEP e0, MEP e1, MEP e2,
        MTP on_plus_side, MTP on_the_other_side)
{
  MFP mf;
  static int id = 0;
  
  mf = NEWP_OBJ (MF);
  if (!mf) return NULL;
  mf->id    = ++id;
  mf->me[0] = e0; me_attach_mf (mf->me[0], mf);
  mf->me[1] = e1; me_attach_mf (mf->me[1], mf);
  mf->me[2] = e2; me_attach_mf (mf->me[2], mf);
  mf->mt[0] = on_plus_side;
  mf->mt[1] = on_the_other_side;
  
  return mf;
}

static void
mv_attach_me (MVP v, MEP e)
{
  if (!v->mel) v->mel = make_list ();
  add_to_tail (v->mel, e);
}

static MEP
new_me (MVP sv, MVP ev, MVP mv)
{
  MEP me;
  static int id = 0;

  me = NEWP_OBJ (ME);
  if (!me) return NULL;
  me->id    = ++id;
  me->mv[0] = sv; mv_attach_me (sv, me);
  me->mv[1] = ev; mv_attach_me (ev, me);
  me->mv[2] = mv;
  me->mfl   = make_list ();
  me->ap    = NULL;
  return me;
}


#define ORDER_3_NUMS(i0, i1, i2, o0, o1, o2)    \
   {                                            \
     int _tmp;                                  \
     if (i0 < i1) {                             \
       o0 = i0;                                 \
       o1 = i1;                                 \
     } else {                                   \
       o0 = i1;                                 \
       o1 = i0;                                 \
     }                                          \
     if (i2 > o1) {                             \
       o2 = i2;                                 \
     } else if (i2 > o0) {                      \
       _tmp = o1;                               \
       o1 = i2;                                 \
       o2 = _tmp;                               \
     } else {                                   \
       o2 = o1;                                 \
       o1 = o0;                                 \
       o0 = i2;                                 \
     }                                          \
   }

#define ORDER_2_NUMS(i0, i1, o0, o1)            \
   {                                            \
     if (i0 < i1) {                             \
       o0 = i0;                                 \
       o1 = i1;                                 \
     } else {                                   \
       o0 = i1;                                 \
       o1 = i0;                                 \
     }                                          \
   }


typedef struct adj_info_t {
  int      face;
  MTP      mt;
}              adj_info_t;



static int
common_face (MTP mt, int fnum, MTP amt, int *afnum)
{
  int af;
  MVP n0, n1, n2, an0, an1, an2;

  n0 = mt->mv[CV_ON_F (fnum, 0)];
  n1 = mt->mv[CV_ON_F (fnum, 1)];
  n2 = mt->mv[CV_ON_F (fnum, 2)];
  FOR (af, 4) {
    an0 = amt->mv[CV_ON_F (af, 0)];
    an1 = amt->mv[CV_ON_F (af, 1)];
    an2 = amt->mv[CV_ON_F (af, 2)];
    if (   (n0 == an0 || n0 == an1 || n0 == an2)
        && (n1 == an0 || n1 == an1 || n1 == an2)
        && (n2 == an0 || n2 == an1 || n2 == an2)
        ) {
      *afnum = af;
      return TRUE;
    }
  }
  return FALSE;
}



static int
add_adj (MESHP st, MTP mt, int fnum)
{
  adj_info_t *i;
  int i0, i1, i2, o0, o1, o2;
  LIST l;

  i = NEWP_OBJ (adj_info_t);
  if (!i) return FALSE;
  i->mt   = mt;
  i->face = fnum;
  i0      = mt->mv[CV_ON_F (fnum, 0)]->id;
  i1      = mt->mv[CV_ON_F (fnum, 1)]->id;
  i2      = mt->mv[CV_ON_F (fnum, 2)]->id;
  ORDER_3_NUMS (i0, i1, i2, o0, o1, o2);
  l = dict_find_obj_with_int_name (st->adjldict, o0);
  if (!l) {
    l = make_list ();
    dict_add_obj_int_name (st->adjldict, l, o0);
  }
  add_to_tail (l, i);
  return TRUE;
}


/*
   Create a new triangulation.
 */
MESHP /* triangulation */
tet_new_triangulation (
                         tet_error_notify_func error_notify
                         )
{
  MESHP st;

  st = NEWP_OBJ (MESH);
  if (!st) return NULL;

  st->geom_degree = 1;          /* linear */
  st->mvdict      = dict_create_dictionary ();
  st->mtdict      = dict_create_dictionary ();
  st->num_mt      = 0;
  st->adjldict    = dict_create_dictionary ();
  st->mfl         = make_list ();
  st->mel         = make_list ();
  st->invalid_rid = 0;
  st->max_node_id = 0;
  st->orientable  = TRUE;

  return st;
}


static MVP
common_mv (MEP e, MEP oe)
{
  if      ((e->mv[0] == oe->mv[0]) || (e->mv[0] == oe->mv[1]))
    return e->mv[0];
  else if ((e->mv[1] == oe->mv[0]) || (e->mv[1] == oe->mv[1]))
    return e->mv[1];
  else
    return NULL;
}

static int
mf_mv (MFP mf, MVP mv[3])
{
  mv[0] = common_mv (mf->me[0], mf->me[1]);
  mv[1] = common_mv (mf->me[1], mf->me[2]);
  mv[2] = common_mv (mf->me[2], mf->me[0]);
  mv[3] = mf->me[1]->mv[2];
  mv[4] = mf->me[2]->mv[2];
  mv[5] = mf->me[0]->mv[2];
  if (mv[3] && mv[4] && mv[5]) return 6;
  else                         return 3;
}

static MVP
find_mv (MESHP st, int id)
{
  return dict_find_obj_with_int_name (st->mvdict, id);
}

static void
attach_mv (MESHP st, MVP mv)
{
  st->max_node_id = max (st->max_node_id, mv->id);
  dict_add_obj_int_name (st->mvdict, mv, mv->id);
}

static void
attach_me (MESHP st, MEP me)
{
  add_to_tail (st->mel, me);
}

static void
attach_mf (MESHP st, MFP mf)
{
  add_to_tail (st->mfl, mf);
}

static void
attach_mt (MESHP st, MTP mt)
{
  st->num_mt++;
  dict_add_obj_int_name (st->mtdict, mt, mt->id);
}




static void
conn_adj_mt (MTP mt, int fnum, MTP amt, int afnum)
{
  mt->adjmt[fnum]   = amt;
  amt->adjmt[afnum] = mt;
}

static MVP
new_mv (int id, LOC *l)
{
  MV *v;
  
  v = NEWP_OBJ (MV);
  if (!v) return NULL;
  v->id    = id;
  v->loc   = *l;
  v->mel   = NULL;
  v->ap    = NULL;
  return v;
}


static MTP
new_mt (int id, MVP mv[10], int rid)
{
  MTP t;
  int i;
  
  t = NEWP_OBJ (MT);
  if (!t) return NULL;
  t->id       = id;
  t->rid      = rid;
  FOR (i, 10) {
    t->mv[i]    = mv[i];
  }
  FOR (i, 4) {
    t->adjmt[i] = NULL;
  }
  t->ap = NULL;
  return t;
}


static void
adj_mt (MESHP st, MTP mt, int fnum, MTP *amt, int *afnum)
{
  adj_info_t *i;
  int i0, i1, i2, o0, o1, o2, af;
  LIST l;
  
  i0      = mt->mv[CV_ON_F (fnum, 0)]->id;
  i1      = mt->mv[CV_ON_F (fnum, 1)]->id;
  i2      = mt->mv[CV_ON_F (fnum, 2)]->id;
  ORDER_3_NUMS (i0, i1, i2, o0, o1, o2);
  l = dict_find_obj_with_int_name (st->adjldict, o0);
  if (l) {
    i = NULL;
    while ((i = (adj_info_t *) get_list_next (l, i))) {
      if (i->mt != mt) {
        if (common_face (mt, fnum, i->mt, &af)) {
         *amt   = i->mt;
         *afnum = af;
         return;
        }
      }
    }
  }
  *amt = NULL;
  *afnum = -1;
}


/*
   Add a node.
 */
int
tet_add_node (MESHP st,
              int id,
              LOC *l
              )
{
  MVP mv;
  mv = find_mv (st, id);
  if (mv) return FALSE;
  mv = new_mv (id, l);
  attach_mv (st, mv);
  return TRUE;
}


/*
   Add a tetrahedron.
   */
int /* TRUE if successfully merged; FALSE otherwise */
tet_add_tet (MESHP st,
             int id,
             int mvids[10],
             int regid
             )
{
  int fnum, afnum;
  MVP mv[10];
  MTP mt;
  MTP amt;
#define VERTS()                                         \
  {                                                     \
    int i, got = 0;                                     \
    FOR (i, 10) {                                       \
      mv[i] = NULL;                                     \
      if (mvids[i] >= 0) {                              \
        got++;                                          \
        mv[i] = find_mv (st, mvids[i]);                 \
        if (!mv[i]) {                                   \
          REPERR (st, "no such node");                  \
          return FALSE;                                 \
        }                                               \
      } else if (i < 4) {                               \
        REPERR (st, "invalid id for a corner node");    \
        return FALSE;                                   \
      } else if (i >= 4 && st->geom_degree == 2) {      \
        REPERR (st, "invalid id for a mid-edge node");  \
        return FALSE;                                   \
      }                                                 \
    }                                                   \
    if (got == 10) st->geom_degree = 2;                 \
  }

  if (regid <= st->invalid_rid) {
    REPERR (st, "Region id should be positive integer");
    return FALSE;
  }
  
  VERTS ();
  mt = new_mt (id, mv, regid); attach_mt (st, mt);

  FOR (fnum, 4) {
    adj_mt (st, mt, fnum, &amt, &afnum);
    if (amt) {
      conn_adj_mt (mt, fnum, amt, afnum);
    } else {
      if (!add_adj (st, mt, fnum)) {
        REPERR (st, "no memory");
        return FALSE;
      }
    }
  }

  st->invalid_rid = min (st->invalid_rid, regid-1);
  
  return TRUE;
}


static ME *
conn_e (MV *v, MV *nv)
{
  ME *e;
  if (!v->mel) return NULL;
  loop_through_list (v->mel, e, ME *) {
    if      (e->mv[0] == v && e->mv[1] == nv)
      return e;
    else if (e->mv[0] == nv && e->mv[1] == v)
      return e;
  }
  return NULL;
}

static MEP
find_me (MESHP st, MVP mv0, MVP mv1)
{
  return conn_e (mv0, mv1);
}

static MEP
get_me (MESHP st, MVP mv0, MVP mv1, MVP mv2)
{
  MEP me;
  me = find_me (st, mv0, mv1);
  if (!me) {
    me = new_me (mv0, mv1, mv2);
    attach_me (st, me);
  }
  return me;
}


static void
orient_mf (MFP mf)
{
  if (mf->mt[0]->rid < mf->mt[1]->rid) {
    /* orient face such that the region id
       on the minus side is smaller than that
       on the plus side */
    MTP mt = mf->mt[0];
    MEP me = mf->me[1];
    mf->mt[0] = mf->mt[1]; mf->mt[1] = mt;
    mf->me[1] = mf->me[2]; mf->me[2] = me;
  }
}

static void
make_surf (MESHP st)
{
  dict_iterator diter;
  int i;
  MFP mf;
  MTP mt;
#define IS_EXTSURF(mt, i) (mt->adjmt[i] == NULL)
#define IS_INTSURF(mt, i) (mt->adjmt[i] && (mt->adjmt[i]->rid != mt->rid))
  
  diter = dict_get_iterator ();
  dict_reset_iterator (diter);
  while ((mt = dict_iter_objs (st->mtdict, diter))) {
    FOR (i, 4) {
      if        (IS_EXTSURF (mt, i)) {
        MEP me0, me1, me2;
        me0 = get_me (st, mt->mv[CV_ON_F (i, 0)], mt->mv[CV_ON_F (i, 1)], mt->mv[MEV_ON_F (i, 1)]);
        me1 = get_me (st, mt->mv[CV_ON_F (i, 1)], mt->mv[CV_ON_F (i, 2)], mt->mv[MEV_ON_F (i, 2)]);
        me2 = get_me (st, mt->mv[CV_ON_F (i, 2)], mt->mv[CV_ON_F (i, 0)], mt->mv[MEV_ON_F (i, 0)]);
        mf = new_mf (me0, me1, me2, NULL, mt);
        attach_mf (st, mf);
      } else if (IS_INTSURF (mt, i)) {
        MEP me0, me1, me2;
        me0 = get_me (st, mt->mv[CV_ON_F (i, 0)], mt->mv[CV_ON_F (i, 1)], mt->mv[MEV_ON_F (i, 1)]);
        me1 = get_me (st, mt->mv[CV_ON_F (i, 1)], mt->mv[CV_ON_F (i, 2)], mt->mv[MEV_ON_F (i, 2)]);
        me2 = get_me (st, mt->mv[CV_ON_F (i, 2)], mt->mv[CV_ON_F (i, 0)], mt->mv[MEV_ON_F (i, 0)]);
        if (!find_mf (st, me0, me1, me2)) {
          mf = new_mf (me0, me1, me2, mt->adjmt[i], mt);
          attach_mf (st, mf);
          orient_mf (mf);
        }
      }
    }
  }
  dict_del_iterator (diter);
}

static void
start_new_region (MESHP mesh, MTP mt)
{
  int rid = mesh->curr_rid;
  MTP smt = mt;
  int i;
  static STACK s = NULL;

  if (!s) s = make_stack ();
  push_stack (s, smt);
  while (!stack_empty (s)) {
    mt = (MTP) pop_stack (s);
    mt->new_rid = rid;
    FOR (i, 4) {
      if (mt->adjmt[i]) {
        if (mt->adjmt[i]->new_rid == mesh->invalid_rid) {
          mt->adjmt[i]->new_rid = rid;
          push_stack (s, mt->adjmt[i]);
        }
      }
    }
  }
}

static void
separate (MESHP mesh)
{
  MTP mt;
  int done;
  dict_iterator diter = dict_get_iterator ();
  
  /* invalidate the new region ids */
  dict_reset_iterator (diter);
  while ((mt = dict_iter_objs (mesh->mtdict, diter))) {
    mt->new_rid = mesh->invalid_rid;
  }

  /* flood the body using the connectivities; */
  mesh->curr_rid = 0;
  while (1) {
    done = TRUE;
    dict_reset_iterator (diter);
    while ((mt = dict_iter_objs (mesh->mtdict, diter))) {
      if (mt->new_rid == mesh->invalid_rid) {
        mesh->curr_rid++;
        start_new_region (mesh, mt);
        done = FALSE;
        break;
      }
    }
    if (done)
      break;
  }

  { /* swap the region identifiers */
    int tmprid;
    dict_reset_iterator (diter);
    while ((mt = dict_iter_objs (mesh->mtdict, diter))) {
      tmprid = mt->rid; mt->rid = mt->new_rid; mt->new_rid = tmprid;
    }
  }
  
  dict_del_iterator (diter);
}


typedef struct shared_edge_t {
  MVP othermv;
  MVP midedgemv;
}              shared_edge_t;


static void
build_shared_edges (MESHP mesh, dict_dictionary shared_edges)
{
  MTP mt;
  int e;
  MVP fmv, smv;
  shared_edge_t *se;
  LIST l;
  LOC loc;
  
  mt = NULL;
  while ((mt = tet_next_mt (mesh, mt))) {
    FOR (e, 6) {
      fmv = mt->mv[g3dtet10_vertex_on_edge[e][0]];
      smv = mt->mv[g3dtet10_vertex_on_edge[e][2]];
      if (smv->id < fmv->id) { MVP amv = fmv; fmv = smv; smv = amv; }
      l = dict_find_obj_with_handle_name (shared_edges, fmv);
      if (!l) {
        l = make_list ();
        dict_add_obj_handle_name (shared_edges, l, fmv);
      } 
      loop_through_list (l, se, shared_edge_t *) {
        if (se->othermv == smv) {
          goto found;
        }
      }
      se = NEWP_OBJ (shared_edge_t);
      se->othermv   = smv;
      VECMAC_3_VEC (loc, =, 0.5 * (fmv->loc), +, 0.5 * (smv->loc));
      se->midedgemv = new_mv (++(mesh->max_node_id), &loc);
      attach_mv (mesh, se->midedgemv);
      add_to_tail (l, se);
    found:
      assert (se);
      assert (se->midedgemv);
      mt->mv[g3dtet10_vertex_on_edge[e][1]] = se->midedgemv;
    }
  }
}

static void
convert_to_quadratic_geom (MESHP mesh)
{
  MEP me;
  MVP fmv, smv;
  LIST l;
  shared_edge_t *se;
  dict_dictionary shared_edges = NULL;

  shared_edges = dict_create_dictionary ();

  build_shared_edges (mesh, shared_edges);
  
  me = NULL;
  while ((me = (MEP) get_list_next (mesh->mel, me))) {
    if (me->mv[2]) {
      REPERR (mesh, "Edge already quadratic");
    }
    fmv = me->mv[0]; smv = me->mv[1];
    if (smv->id < fmv->id) { MVP amv = fmv; fmv = smv; smv = amv; }
    l = dict_find_obj_with_handle_name (shared_edges, fmv);
    assert (l);
    loop_through_list (l, se, shared_edge_t *) {
      if (se->othermv == smv) {
        goto found;
      }
    }
  found:
    assert (se);
    assert (se->midedgemv);
    me->mv[2] = se->midedgemv;
    attach_mv (mesh, me->mv[2]);
  }

  { /* clean up */
    shared_edge_t *se;
    LIST l;
    dict_iterator diter = dict_get_iterator ();
    dict_reset_iterator (diter);
    while ((l = dict_iter_objs (shared_edges, diter))) {
      loop_through_list (l, se, shared_edge_t *) {
        NEWP_FREE_OBJ (se);
      }
      free_list (l, NOT_NODES);
    }
    dict_destroy_dictionary (shared_edges);
    dict_del_iterator (diter);
  }
}

/*
   Prepare triangulation for topology extraction.
 */
int
tet_prepare (MESHP mesh,
             int separate_unconnected_tets, /* should separation of unconnected tet groups
                                               into separate regions be attempted?
                                             */
             int make_quadratic /* should the geometry be completed to be quadratic
                                 (by generating mid-edge nodes)? */
             )
{
  /* not needed any longer */
  dict_destroy_dictionary (mesh->adjldict);
  mesh->adjldict = NULL;

  make_surf (mesh);

  if (separate_unconnected_tets)
    separate (mesh);

  if (make_quadratic && mesh->geom_degree == 1)
    convert_to_quadratic_geom (mesh);

  return TRUE;
}

/*
   Iterate surface faces.
 */
MFP /* NULL if there are no more triangles */
tet_next_mf (MESHP st, /* surface triangulation */
              MFP o /* start iteration by passing NULL */
              )
{
  return (MFP) get_list_next (st->mfl, o);
}

/*
   Iterate edges.
 */
MEP /* NULL if there are no more edges */
tet_next_me (MESHP st, /* surface triangulation */
              MEP o /* start iteration by passing NULL */
              )
{
  return (MEP) get_list_next (st->mel, o);
}

/*
   Iterate vertices.
 */
MVP /* NULL if there are no more vertices */
tet_next_mv (MESHP st, /* surface triangulation */
              MVP o /* start iteration by passing NULL */
              )
{
  MVP amv;
  static dict_iterator diter = NULL;
  if (!diter) diter = dict_get_iterator ();
  if (!o) dict_reset_iterator (diter);
  while ((amv = dict_iter_objs (st->mvdict, diter))) {
    if (amv->mel) /* is a surface mv? */
      return amv;
  }
  return NULL;
}


/*
   Iterate internal vertices (vertices not on the external surface).
 */
MVP /* NULL if there are no more vertices */
tet_next_intmv (MESHP st, /* surface triangulation */
                MVP o /* start iteration by passing NULL */
                )
{
  MVP amv;
  static dict_iterator diter = NULL;
  if (!diter) diter = dict_get_iterator ();
  if (!o) dict_reset_iterator (diter);
  while ((amv = dict_iter_objs (st->mvdict, diter))) {
    if (!amv->mel) /* is interior mv? */
      return amv;
  }
  return NULL;
}


/*
   Get triangle normal.
 */
int /* TRUE if normal is valid; FALSE otherwise */
tet_mf_get_normal (MFP mf, /* triangle */
                   double xi, double eta,
                   LOC *normal
                   )
{
  LOC n = {0,0,0}, f, s;
  MVP mv[3];
  double nn;
  mf_mv (mf, mv);
  VECMAC_3_VEC (f, =, (mv[1]->loc), -, (mv[0]->loc));
  VECMAC_3_VEC (s, =, (mv[2]->loc), -, (mv[0]->loc));
  VECMAC_CROSS (n, =, f, x, s);
  nn = VECMAC_NORM (n);
  if (nn <= 0) return FALSE;
  VECMAC_NORMALIZE (n);
  VECMAC_2_VEC ((*normal), =, n);
  return TRUE;
}



static void
destroy_mv (MESHP st, MV *v)
{
  NEWP_FREE_OBJ (v);
}



static void
destroy_mt (MESHP st, MTP t)
{
  NEWP_FREE_OBJ (t);
}

/*
   Destroy surface mesh.
 */
void
tet_destroy_triangulation (MESHP st
                             )
{
  MVP v;
  MTP t;
  dict_iterator diter;

  diter = dict_get_iterator ();

  dict_reset_iterator (diter);
  while ((v = dict_iter_objs (st->mvdict, diter))) {
    destroy_mv (st, v);
  }
  dict_reset_iterator (diter);
  while ((t = dict_iter_objs (st->mtdict, diter))) {
    destroy_mt (st, t);
  }
  
  NEWP_FREE_OBJ (st);
  dict_del_iterator (diter);
}



/*
   Get edges.
 */
void
tet_mf_get_me (MFP t, /* triangle */
               MEP e[3] /* edges */
               )
{
  int i;
  FOR (i, 3) { e[i] = t->me[i]; }
}

/*
   Get associated pointer.
 */
void * /* pointer stored by the triangle */
tet_mf_get_ap (MFP mf /* triangle */
                )
{
  return mf->ap;
}

/*
   Set associated pointer.
 */
void * /* previously stored pointer */
tet_mf_set_ap (MFP t, /* triangle */
                void *ap /* pointer */
                )
{
  void *oap = t->ap;
  t->ap = ap;
  return oap;
}

/*
   Get surface tag.
 */
void * /* pointer stored by the triangle */
tet_mf_get_surf (MFP mf /* triangle */
                )
{
  return NULL; /* don't have any other indications
                  than those coming from the tet mesh*/
}


/*
   Get id of mesh edge.
 */
int /* edge id */
tet_me_get_id (MEP e /* edge */
                )
{
  return e->id;
}

/*
   Get associated pointer.
 */
void *
tet_me_get_ap (MEP e)
{
  return e->ap;
}

/*
   Get incident vertices.
 */
int /* returns number of vertices */
tet_me_get_mv (MEP e,
               MVP v[3] /* vertices in the order start, end, middle (if present) */
               )
{
  v[0] = e->mv[0];
  v[1] = e->mv[1];
  v[2] = e->mv[2];
  if (v[2]) return 3;
  else      return 2;
}

/*
   Get incident vertices.
 */
int
tet_mt_get_mv (MTP t, MVP v[10])
{
  int i;
  FOR (i, 10) { v[i] = t->mv[i]; }
  if (v[4] && v[5] && v[6] && v[7] && v[8] && v[9]) return 10;
  else                                              return  4;
}

int
tet_mt_get_id (MTP mt)
{
  return mt->id;
}

/*
   Get region.
 */
MRID
tet_mt_get_region (MTP mt)
{
  return (MRID) mt->rid;
}

/*
   Get associated pointer.
 */
void *
tet_mt_get_ap (MTP t)
{
  return t->ap;
}

/*
   Set associated pointer.
 */
void *
tet_mt_set_ap (MTP e,
                void *ap
                )
{ 
  void *oap = e->ap;
  e->ap = ap;
  return oap;
}

/*
   Iterate triangles incident to edge.
 */
MFP /* next triangle or NULL if there are no more */
tet_me_get_next_mf (MEP me,
                    MFP mf
                    )
{
  return (MFP) get_list_next (me->mfl, mf);
}


/*
   Count incident triangles at edge.
 */
int
tet_me_get_num_inc_mf (MEP e)
{
  return list_length (e->mfl);
}

/*
   Set associated pointer.
 */
void *
tet_me_set_ap (MEP e,
                void *ap
                )
{ 
  void *oap = e->ap;
  e->ap = ap;
  return oap;
}

/*
   Get associated pointer.
 */
void *
tet_mv_get_ap (MVP v /* vertex */
                )
{
  return v->ap;
}

/*
   Set associated pointer.
 */
void *
tet_mv_set_ap (MVP v, void *ap)
{
  void *oap = v->ap;
  v->ap = ap;
  return oap;
}


/*
   Get vertex id
 */
int
tet_mv_get_id (MVP v)
{
  return v->id;
}

/*
   Get vertex location.
 */
void
tet_mv_get_loc (MVP v, /* vertex */
                 LOC *p /* location */
                 )
{
  *p = v->loc;
}

/*
   Iterate edges incident to vertex
 */
MEP
tet_mv_get_next_me (MVP v, MEP e)
{
  if (!v->mel)  return NULL;
  else          return (MEP) get_list_next (v->mel, e);
}


/*
   Get triangle id
 */
int
tet_mf_get_id (MFP t)
{
  return t->id;
}

/*
   Get incident vertices.
 */
int
tet_mf_get_mv (MFP t, /* triangle */
               MVP v[6] /* vertices in ccw order when
                         looking against the normal;
                         vertex 0 is shared by edges 0 and 1,
                         vertex 1 is shared by edges 1 and 2,
                         vertex 2 is shared by edges 2 and 0,
                         vertex 3 is the mid-side node of edge 1,
                                 vertex 4 is the mid-side node of edge 2,
                                 vertex 5 is the mid-side node of edge 0 (the last three
                                 only for the quadratic geometry)
                         */
               )
{
  return mf_mv (t, v);
}

MTP
tet_find_mt_with_id (MESHP st, int id)
{
  return dict_find_obj_with_int_name (st->mtdict, id);
}



/*
   Get regions on plus and minus side (plus side is where the normal points).
 */
int
tet_mf_get_regions (MFP mf, MRID *mridplus, MRID *mridminus)
{
  int retcode = 0;
  *mridplus = NULL;
  if (mf->mt[0]) {
    *mridplus = (MRID) mf->mt[0]->rid;
    retcode++;
  }
  *mridminus = NULL;
  if (mf->mt[1]) {
    *mridminus = (MRID) mf->mt[1]->rid;
    retcode++;
  }
  return retcode;
}

int
tet_total_of_tets (MESHP mesh)
{
  int n = 0;
  MTP mt;
  dict_iterator diter = dict_get_iterator ();
  dict_reset_iterator (diter);
  while ((mt = dict_iter_objs (mesh->mtdict, diter))) {
    n++;
  }
  dict_del_iterator (diter);
  return n;
}

int
tet_total_of_nodes (MESHP mesh)
{
  int n = 0;
  MVP mv;
  dict_iterator diter = dict_get_iterator ();
  dict_reset_iterator (diter);
  while ((mv = dict_iter_objs (mesh->mvdict, diter))) {
    n++;
  }
  dict_del_iterator (diter);
  return n;
}

int
tet_total_of_surf_triangles (MESHP mesh)
{
  return list_length (mesh->mfl);
}


int
tet_total_of_surf_edges (MESHP mesh)
{
  return list_length (mesh->mel);
}


/*
   Get tangent vector to an edge at a vertex.
 */
void
tet_me_get_tangent (MEP me, MVP mv, LOC *tangent)
{
  VECMAC_3_VEC ((*tangent),  =, me->mv[1]->loc, -, me->mv[0]->loc);
  VECMAC_NORMALIZE ((*tangent));
}


int
tet_geom_degree (MESHP mesh) {
  return mesh->geom_degree;
}


MTP
tet_next_mt (MESHP mesh, MTP mt)
{
  MTP omt;
  static dict_iterator diter = NULL;
  if (!diter) diter = dict_get_iterator ();
  if (!mt) { dict_reset_iterator (diter); }
  omt = dict_iter_objs (mesh->mtdict, diter);
  return omt;
}

int
tet_num_mt (MESHP mesh) { return mesh->num_mt; }

double
tet_vltol (MESHP mesh)
{
  return 0;                     /* no tolerance applies */
}

int
tet_is_orientable (MESHP mesh) { return mesh->orientable; }
